/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file alarm.c
 */

#include <stdio.h>
#include <stdarg.h>

#include "kernel.h"

#ifndef __GNUC__
#pragma -s1
#endif

#include "main.h"
#include "canvas.h"
#include "icons.h"

#define MAX_ALARMS (64)

typedef void (* alarm_proc)(int, int);

typedef struct {
  alarm_proc p;
  int time;
  int h;
  int used;
} alarm;

static alarm alarms[MAX_ALARMS];
static int next_time, in_alarm = 0;

static void __alarm(void)
{
  int i, time = main_time(), delta_t = 1 << 30, delta_i = -1;

  in_alarm = 1;

  /* Scan through alarms, looking for ones that should be triggered */
  for(i = 0; i < MAX_ALARMS; i ++) {
    if(alarms[i].used) {
      if(alarms[i].time - time <= 0) {
	alarms[i].used = 0;
	alarms[i].p(time, alarms[i].h);
      }
    }
  }

  /* Find the next alarm to go off */
  for(i = 0; i < MAX_ALARMS; i ++) {
    if(alarms[i].used) {
      if(alarms[i].time - time < delta_t) {
	delta_t = alarms[i].time - time;
	delta_i = i;
      }
    }
  }

  /* If an alarm is scheduled, tell the main core about it, otherwise, don't bother */
  if(delta_i == -1)
    main_nextnull(next_time = 0);
  else
    main_nextnull(next_time = time + delta_t);

  in_alarm = 0;
}

void alarm_init(void)
{
  int i;

  /* Clear all the alarms */
  for(i = 0; i < MAX_ALARMS; i ++)
    alarms[i].used = alarms[i].time = 0;

  /* Register an alarm handler */
  main_addhandler(HANDLER_NULL, __alarm);
}

void alarm_set(int time, alarm_proc p, int h)
{
  int i;

  for(i = 0; i < MAX_ALARMS; i ++) {
    if(!alarms[i].used) {
      alarms[i].used = 1;
      alarms[i].time = time;
      alarms[i].p = p;
      alarms[i].h = h;
      if(!in_alarm) {
	if((next_time - time > 0) || (!next_time))
	  next_time = time;
	main_nextnull(next_time);
      }
      return;
    }
  }
  exit(simple_warning("Too many alarms set", 1));
}

void alarm_cancel(int h)
{
  int i, delta_t = 1 << 30, delta_i = -1, time = main_time();

  for(i = 0; i < MAX_ALARMS; i ++) {
    if(alarms[i].used && (alarms[i].h == h)) {
      alarms[i].used = 0;
    } else {
      if(alarms[i].used) {
	if(alarms[i].time - time < delta_t) {
	  delta_t = alarms[i].time - time;
	  delta_i = i;
	}
      }
    }
  }

  /* If we weren't called in the middle of an alarm, schedule the next one */
  if(!in_alarm) {
    if(delta_i == -1) {
      main_nextnull(next_time = 0);
    } else {
      main_nextnull(next_time = time + delta_t);
    }
  }
}
